#!/usr/bin/env python
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Based on code from:
#   PyODConverter (Python OpenDocument Converter) v1.0.0 - 2008-05-05
#   Copyright (C) 2008 Mirko Nasato <mirko@artofsolving.com>
#   Licensed under the GNU LGPL v2.1 - or any later version.
#   http://www.gnu.org/licenses/lgpl-2.1.html
#
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

################################################################################
# TODO LIST:
# 1) aggiungere la stampa del report o riepilogo strutturato (se richiesto dall'utente)

################################################################################
# IMPORT BASIC MODULES #########################################################
################################################################################
import sys
import os
import subprocess
import time
################################################################################
# IMPOSTAZIONI LOGGINGS ########################################################
################################################################################
import logging
logger = logging.getLogger('preventares')
################################################################################
# IMPORT OPENOFFICE MODULES ####################################################
################################################################################
import uno
import openoffice.interact
from com.sun.star.awt import Size
from com.sun.star.beans import PropertyValue
from com.sun.star.connection import NoConnectException

################################################################################
# COSTANTI DI OPENOFFICE #######################################################
################################################################################
_ControlCharacter = "com.sun.star.text.ControlCharacter"
_TextContentAnchorType = "com.sun.star.text.TextContentAnchorType"
_ParagraphAdjust = "com.sun.star.style.ParagraphAdjust"
PARAGRAPH_BREAK  = uno.getConstantByName(_ControlCharacter+ ".PARAGRAPH_BREAK")
LINE_BREAK       = uno.getConstantByName(_ControlCharacter+ ".LINE_BREAK")
HARD_HYPHEN      = uno.getConstantByName(_ControlCharacter+ ".HARD_HYPHEN")
SOFT_HYPHEN      = uno.getConstantByName(_ControlCharacter+ ".SOFT_HYPHEN")
HARD_SPACE       = uno.getConstantByName(_ControlCharacter+ ".HARD_SPACE")
APPEND_PARAGRAPH = uno.getConstantByName(_ControlCharacter+ ".APPEND_PARAGRAPH")
AS_CHARACTER     = uno.getConstantByName(_TextContentAnchorType+ ".AS_CHARACTER")
# ParagraphAdjust: These values describe the formatting of a text paragraph. 
LEFT        = uno.getConstantByName(_ParagraphAdjust+ ".LEFT")
RIGHT       = uno.getConstantByName(_ParagraphAdjust+ ".RIGHT")
BLOCK       = uno.getConstantByName(_ParagraphAdjust+ ".BLOCK")
CENTER      = uno.getConstantByName(_ParagraphAdjust+ ".CENTER")
STRETCH     = uno.getConstantByName(_ParagraphAdjust+ ".STRETCH")
# costanti colori base
COLOR_BLACK = 0
COLOR_BLUE  = 255

################################################################################
# CLASSE PER AVVIARE UNA NUOVA ISTANZA DI OPENOFFICE ###########################
################################################################################
class OpenOfficeRunner:
    """A class to open a new openoffice instance of writer"""
    def __init__ (self, *args):
        # standard linux installation ooffice paths
        self.__executable=['/usr/bin/ooffice', '/usr/bin/libreoffice']
        self.ooffice = None
        self.exec_filename = None
        if len(args) > 0:
            self.__executable.extend(args)
    
    def run(self):
        for filename in self.__executable:
            if os.path.exists(filename):
                self.exec_filename = filename
                try:
                   self.ooffice = subprocess.Popen([filename, 
                               "--accept=socket,host=localhost,port=2002;urp;", 
                               "--nologo"])
                except:
                   logger.error("""Avviamento di openoffice per collegamento
                         a socket fallito. Avvio openoffice in modalita
                         normale.""")
                   self.ooffice = subprocess.Popen([filename, 
                                  "--writer", '--nologo'])
                # attendi 1 sec per permettere l'avviamento del programma
                time.sleep(1)
                break
        if self.ooffice is None:
            logger.error("Error OpenOffice executable not found.")
            sys.exit(0)
        # get the central desktop object
        desktop = openoffice.interact.Desktop()
        return desktop

################################################################################
# CLASSE CON FUNZIONI PER SCRIVERE SU UN DOCUMENTO OPENOFFICE WRITER ###########
################################################################################
class PrintToWriter:
    """
    A class which allows conveniently printing stuff into a Writer
    document.
    
    Very useful for debugging, general output purposes, or even for
    creating reports.
    """
    
    def __init__(self, desktop=None, fontName="Sans", fontSize=8,
                 odt_model='preventares-cm.ott'):
        if not desktop:
            desktop = openoffice.interact.Desktop() # TODO: defaultDesktop? 
            #target='_blank', hidden=1) FIXME apre una nuova istanza di 'writer'
        self.odt_model = odt_model
        try:
            self._doc = desktop.openFile(self.odt_model)
            logger.info("Apertura del file modello '{0}'".format(self.odt_model))
        except:
            logger.error("File modello '{0}' non trovato. Apertura di un nuovo documento.".format(self.odt_model))
            self._doc = desktop.newDoc('swriter', )
        self._text = self._doc.getText()
        self._cursor = self._text.createTextCursor()
        self._font = fontName
        self._size = fontSize
        self._color = 0
        # imposta lo stile di pagina
        self.setPageStyleProperty()

    def writeControlCharacter(self, char, absorb=False):
        # TODO: what does 'absorb' mean?
        self._text.insertControlCharacter(self._cursor, char, absorb)

    def write(self, string, absorb=False):
        self._text.insertString(self._cursor, str(string), absorb)

    def writelines(self, lines, absorb=False):
        for line in lines:
            self._text.insertString(self._cursor, line, absorb)

    def writeTab(self, absorb=False):
        self.write("\t", absorb)

    def paragraphBreak(self, absorb=False):
        self.writeControlCharacter(PARAGRAPH_BREAK, absorb)

    def writeLn(self, *args):
        # imposta le proprieta del testo per la cella
        self.writelines(args)
        self.paragraphBreak()

    def writeTabSeparated(self, *args):
        if not args:
            return
        arg = args[0]
        self.write(arg)
        for arg in args[1:]:
            self.writeTab()
            self.write(arg)

    def insertTextIntoCell(self, table, cellName, text, font=None, size=None, 
                           color=None, shadow=False, ParaAdjust=LEFT):
        # imposta il font e il carattere della cella
        Tcursor = table.createCursorByCellName(cellName)
        # imposta le proprietà del cursore
        if (font is not None) or (size is not None) or (color is not None) or \
           (shadow is True) or (ParaAdjust != LEFT):
              self.setCursorProperty(cursor=Tcursor, font=font, size=size, 
                                     color=color, shadow=shadow, 
                                     ParaAdjust=ParaAdjust)
        # imposta il testo nella cella
        table.getCellByName( cellName ).setString( text )

    def insertValueIntoCell(self, table, cellName, value, font=None, size=None, 
                            color=None, shadow=False, ParaAdjust=LEFT):
        # imposta il font e il carattere della cella
        Tcursor = table.createCursorByCellName(cellName)
        # imposta le proprietà del cursore
        if (font is not None) or (size is not None) or (color is not None) or \
           (shadow is True) or (ParaAdjust != LEFT):
              self.setCursorProperty(cursor=Tcursor, font=font, size=size, 
                                     color=color, shadow=shadow, 
                                     ParaAdjust=ParaAdjust)
        # imposta il valore numerico della cella
        table.getCellByName(cellName).setValue(value)

    def insertFormulaIntoCell(self, table, cellName, formula, font=None, size=None,
                              color=None, shadow=False, ParaAdjust=LEFT):
        # imposta il font e il carattere della cella
        Tcursor = table.createCursorByCellName(cellName)
        # imposta le proprietà del cursore
        if (font is not None) or (size is not None) or (color is not None) or \
           (shadow is True) or (ParaAdjust != LEFT):
              self.setCursorProperty(cursor=Tcursor, font=font, size=size, 
                                     color=color, shadow=shadow, 
                                     ParaAdjust=ParaAdjust)
        # imposta la formula all'interno della cella (es. "sum <A3:C3>")
        table.getCellByName(cellName).setFormula(formula)

    def createTable(self, rows, columns, backcolor=None, transparent=False,
                    LeftMargin=None, RightMargin=None):
        """creates a new writer document and inserts a table with some data (also known as the SWriter sample)"""
        text = self._doc.Text
        # create a text table
        table = self._doc.createInstance( "com.sun.star.text.TextTable" )
        # with N rows and N columns
        table.initialize (rows, columns)
        # imposta i margini sinistro e destro della tabella
        table.HoriOrient = 0 #'com::sun::star::text::HoriOrientation::NONE
        if LeftMargin is not None: table.LeftMargin = LeftMargin#2000
        if RightMargin is not None: table.RightMargin = RightMargin#1500
        # ...
        text.insertTextContent( self._cursor, table, False )
        #TODO imposta la dimensione delle colonne della tabella
        for a in dir(table):
            if "Column" in a: print a  
        #table.autoFormat
        # imposta la trasparenza e il colore di sfondo della tabella
        table.setPropertyValue( "BackTransparent", transparent )
        if backcolor is not None:
             table.setPropertyValue( "BackColor", backcolor )
        return table

    def setTableHeading(self, table):
        """"""
        # imposta la prima riga
        t_rows = table.Rows
        row = t_rows.getByIndex(0)
        row.setPropertyValue( "BackTransparent", False)
        row.setPropertyValue( "BackColor", 6710932 )
        row.IsAutoHeight = True
        row.IsSplitAllowed = True
        # inserisce l'intestazione delle colonne
        self.insertTextIntoCell( table, "A1", "ART.", ParaAdjust=CENTER)
        self.insertTextIntoCell( table, "B1", "DESCRIZIONE", ParaAdjust=CENTER)
        self.insertTextIntoCell( table, "C1", "U.M.", ParaAdjust=CENTER)
        self.insertTextIntoCell( table, "D1", "QUANTITÀ", ParaAdjust=CENTER)
        self.insertTextIntoCell( table, "E1", "P.Unitario", ParaAdjust=CENTER)
        self.insertTextIntoCell( table, "F1", "P.Totale", ParaAdjust=CENTER)        

    def createFrame(self, size_x=15000, size_y=400, color=None):
        # istruzioni per creare un frame (riquadro) col testo all'interno
        if color is None: color = COLOR_BLACK # colore nero
        textFrame = self._doc.createInstance( "com.sun.star.text.TextFrame" )
        textFrame.setSize( Size(size_x, size_y))
        textFrame.setPropertyValue( "AnchorType" , AS_CHARACTER )
        self._text.insertTextContent( self._cursor, textFrame, 0 )
        return textFrame

    def insertTextIntoFrame(self, textFrame, text):
        textInTextFrame = textFrame.getText()
        # posiziona il cursore alla fine della cella
        tRange = textInTextFrame.End
        tRange.String = text
        #create an XTextRange at the end of the document
        cursorInTextFrame = textInTextFrame.createTextCursor()
        # inserisce il testo nel frame
        textInTextFrame.insertString( cursorInTextFrame, text, 0 )

    def setPageStyleProperty(self, font=None, size=None, color=None, 
                             shadow=False, ParaAdjust=LEFT):
        # set font
        if font is None: font = self._font
        # set font size
        if size is None: size = self._size
        # set the paragraph style
        StyleFamilies = self._doc.getStyleFamilies()
        ParagraphStyles = StyleFamilies.getByName("ParagraphStyles").getByName("Standard")
        ParagraphStyles.CharShadowed = shadow
	    ParagraphStyles.CharColor = self._color
	    ParagraphStyles.CharFontName = font
	    ParagraphStyles.CharHeight = size
	    #imposta la formattazione di pagina 0--> LEFT, 1-->RIGHT, 2-->BLOCK
	    ParagraphStyles.ParaAdjust = ParaAdjust
	    # imposta il colore del carattere
        if color is None:
            ParagraphStyles.setPropertyValue( "CharColor", self._color )
        else: 
            ParagraphStyles.setPropertyValue( "CharColor", color)
	    
	    
    def setCursorProperty(self, cursor=None, font=None, size=None, color=None, 
                          shadow=False, ParaAdjust=LEFT):
        if cursor is None: cursor=self._cursor
        # set font
        if font is None: font = self._font
        cursor.setPropertyValue("CharFontName", font)
        # set font size
        if size is None: size = self._size
        cursor.setPropertyValue("CharHeight", size)
        # imposta il colore del carattere
        if color is None:
            cursor.setPropertyValue( "CharColor", self._color )
        else: cursor.setPropertyValue( "CharColor", color)
        # imposta se il carattere è ombreggiato
        cursor.setPropertyValue( "CharShadowed", shadow )
        # imposta la formattazione di pagina 0--> LEFT, 1-->RIGHT, 2-->BLOCK
        cursor.ParaAdjust = ParaAdjust
        
    def setCharacterProperties(self, cursor=None, font=None, size=None, 
                               color=None, shadow=False):
        """"""
        chrProp = self._doc.createInstance("com.sun.star.style.CharacterProperties")
        #chrProp.setPropertyValue("CharFontStyleName", styleName)
        if cursor is None: cursor=self._cursor
        # set font
        if font is None: font = self._font
        chrProp.setPropertyValue("CharFontName", font)
        # set font size
        if size is None: size = self._size
        chrProp.setPropertyValue("CharHeight", size)
        # imposta il colore del carattere
        if color is None:
            chrProp.setPropertyValue( "CharColor", self._color )
        else: chrProp.setPropertyValue( "CharColor", color)
        # imposta se il carattere è ombreggiato
        chrProp.setPropertyValue( "CharShadowed", shadow )
            
    def getCursorProperty(self, cursor=None):
        if cursor is None: cursor=self._cursor
        return "FontName={0}, Height={1}, Color={2}, Shadowed={3}".format(
                cursor.getPropertyValue("CharFontName"),
                cursor.getPropertyValue("CharHeight"),
                cursor.getPropertyValue("CharColor"),
                cursor.getPropertyValue("CharShadowed"))

    def insertTextIntoHeader(self, text, font=None, size=None, 
                              color=None, shadow=False, imageURL=None):
        """imposta l'intestazione del documento"""
        StyleFamilies = self._doc.getStyleFamilies()
        PageStyle = StyleFamilies.getByName("PageStyles")
        oStyle = StyleFamilies.getByName("PageStyles").getByName("Standard")
	    oStyle.HeaderIsOn = True
	    oStyle.HeaderIsShared = True
	    oStyle.HeaderHeight = 500 # 0.5 cm - header height
	    # set 0.5 cm distance from page text area to header
	    oStyle.HeaderBodyDistance = 500 
	    headerCursor = oStyle.HeaderText.createTextCursor()
	    self.setCursorProperty(cursor=headerCursor, font=font, size=size,
	                           color=color, shadow=shadow, ParaAdjust=CENTER)
	    oStyle.HeaderText.insertString(headerCursor, text, False)
	    # inserimento di immagine
	    if imageURL is not None:
	        GraphicObj = self.insertGraphicObj(imageURL,
	                                           text=oStyle.HeaderText,
	                                           cursor=headerCursor)
	        if GraphicObj.Height+500 > oStyle.HeaderHeight:
	            #set a new Header Height
	            oStyle.HeaderHeight = 500 + GraphicObj.Height
            #if GraphicObj.Width > ...:

    def insertTextIntoFooter(self, text=None, font=None, size=None, 
                              color=None, shadow=False):
        """imposta il piè di pagina del documento"""
        StyleFamilies = self._doc.getStyleFamilies()
        PageStyle = StyleFamilies.getByName("PageStyles")
        oStyle = StyleFamilies.getByName("PageStyles").getByName("Standard")
	    oStyle.FooterIsOn = True
	    oStyle.FooterIsShared = True
	    oStyle.FooterHeight = 500 # 0.5 cm - header height
	    oStyle.FooterBodyDistance = 500 # 0.5 cm - distance from page text area to header
	    footerCursor = oStyle.FooterText.createTextCursor()
	    self.setCursorProperty(cursor=footerCursor, font=font, size=size,
	                           color=color, shadow=shadow, ParaAdjust=CENTER)
	    if text is not None:
	        oStyle.FooterText.insertString(footerCursor, text, False)
	    else:
	        # crea i campi 'numero di pagine' e 'pagina corrente'
	        numType = uno.getConstantByName("com.sun.star.style.NumberingType"+".ARABIC")
	        oField = self._doc.createInstance("com.sun.star.text.TextField.PageCount")
	        oField.NumberingType = numType
	        tField = self._doc.createInstance("com.sun.star.text.TextField.PageNumber")
	        tField.NumberingType = numType
	        oStyle.FooterText.insertString(footerCursor, "Pagina ", False)
	        oStyle.FooterText.insertTextContent(footerCursor, tField, False)
	        oStyle.FooterText.insertString(footerCursor, " di ", False)
	        oStyle.FooterText.insertTextContent(footerCursor, oField, False)  

    def insertGraphicObj(self, imageURL, text=None, cursor=None,
                         height=None, width=None):
        """inserisce un'immagine all'interno del testo ODT"""
        if os.path.exists(imageURL):
            logger.error("URL immagine '{0}' non esistente".format(imageURL))
        if cursor is None: cursor = self._cursor
        if text is None: text = self._text
        # create a text table
        TextGraphicObject = self._doc.createInstance("com.sun.star.text.TextGraphicObject")
        # imposta i margini e le proprieta dell'immagine
        location = uno.getConstantByName("com.sun.star.style.GraphicLocation"+".LEFT_TOP")
        TextGraphicObject.BackGraphicLocation = location
        #TextGraphicObject.ContourOutside = True
        TextGraphicObject.GraphicURL = imageURL
        # impostazione orientamento immagine
        TextGraphicObject.HoriOrient = 0 #left
        TextGraphicObject.VertOrient = 0 #top
        # impostazione dei margini intorno all'immagine
        TextGraphicObject.LeftMargin = 0
        TextGraphicObject.RightMargin = 0
        TextGraphicObject.TopMargin = 0
        TextGraphicObject.BottomMargin = 0
        # impostazione della larghezza e altezza
        if height is not None:
            TextGraphicObject.Height = height
        if width is not None:
            TextGraphicObject.Width = width
        # debug informations
        logger.debug("Image Size: Height: {0}; Width:{1}".format(
                      TextGraphicObject.Height, 
                      TextGraphicObject.Width))
        # inserisci l'immagine nel documento
        text.insertTextContent( cursor, TextGraphicObject, 0 )
        return TextGraphicObject
        

################################################################################
# PREVENTARES CONVERT ##########################################################
################################################################################
class PreventaConvertToWriter:
    """
    A class which allows to output (print) preventares bill-of-cost to openoffice
    or libreoffice writer (.odt) documents.
    """
    
    def __init__(self, preventa, title='',font="Sans", size=8, 
                 output_type='computo', descr_breve=True, descr_estesa=True, 
                 note=True, no_chapter=False, analisi=False, imageURL=None,
                 odt_model=None):
        # inizializzo le costanti
        self.preventa = preventa
        self.output_type = output_type
        self.title = title
        self.descr_breve = descr_breve
        self.descr_estesa = descr_estesa
        self.note = note
        self.no_chapter = no_chapter
        self.analisi = analisi
        self.imageURL = imageURL
        self.desktop = None
        self.nomeDoc = str()
        self.odt_model = odt_model
        # apro il programma 'openoffice writer'
        oo=OpenOfficeRunner()
        self.desktop = oo.run()
        # ricavo la lista degli articoli da stampare
        lista_prv, lista_rpt = self.__preventa_convert()
        # scrivi su openoffice l'output del comando
        self.__write_to_owriter(lista_prv, lista_rpt, font, size, odt_model)
        # informa l'utente o il logger che il processo è terminato con successo
        logger.info("Esportazione eseguita con successo. Programma Terminato.")

    def __preventa_convert(self):
        # inizializza le liste
        lista_prv = list()
        lista_rpt = list()
        if self.output_type == 'computo': 
            func = self.preventa.print_computo
            func_rp = self.preventa.print_report_computo
        elif self.output_type == 'lavorazioni': 
            func = self.preventa.print_lavorazioni
            func_rp = self.preventa.print_report_lavorazioni
        elif self.output_type == 'epu': 
            func = self.preventa.print_epu
            func_rp = None
        #TODO add in other options
        #elif self.output_type == 'analisi': pass 
        #elif self.output_type == 'report-cmp': pass
        #elif self.output_type == 'report-epu': pass
        #elif self.output_type == 'report-inc_mo': pass
        #elif self.output_type == 'report-inc_sic': pass
        else: 
            func = self.preventa.print_computo
            func_rp = self.preventa.print_report_computo
        sup, cat, sub = (None, None, None)
        lista_prv = func(sup, cat, sub, #super, cat, sub
                         self.descr_breve, 
                         self.descr_estesa, 
                         self.note,
                         False, #print_only_if_not_0
                         self.no_chapter, 
                         self.analisi)
        # riepilogo strutturato
        if func_rp is not None:
            lista_rpt = func_rp(sup, cat, sub, 
                                print_only_if_not_0=False, 
                                print_only_grand_total=False)
        # nome del lavoro in stampa
        self.nomeDoc = self.preventa.dati_generali_list()[4]
        return lista_prv, lista_rpt
        
    def __write_to_owriter(self, lista_prv, lista_rpt, font, size, odt_model):
        # funzioni di conversione float --> str
        def convert_to_str(n):
           if type(n) is not float: return str(n)
           return round(n, 2)
        # apro il documento openoffice
        doc = PrintToWriter(desktop=self.desktop, 
                            fontName=font, 
                            fontSize=size,
                            odt_model=odt_model)
        # impostazione del tipo di testo (colore, font, ecc.)
        # imposta l'intestazione e piè di pagina del documento
        H_txt = "{0} {1}".format(self.title, self.nomeDoc)
        doc.insertTextIntoHeader(H_txt, imageURL=self.imageURL)
        doc.insertTextIntoFooter(text=None)
        # creo una tabella per contenere gli articoli
        rows = len(lista_prv)+1
        columns = 6
        table = doc.createTable(rows, columns)
        # crea le colonne di intestazione
        #doc.setTableHeading(table)
        # parsing degli elementi della lista e li dispongo nella tabella
        i=2
        for art in lista_prv:
            cod, tar, desc, um, quant, pu, pt, a, b, c, d = art
            nome, desc_e, desc_b, note = desc
            # inserimento di carattere speciale prima della stampa
            if cod.startswith('$$$'): pass #FIXME testo in grassetto!!
            # preparazione del testo prima della stampa
            text = "{0}\n{1}\n{2}\n{3}".format(nome, desc_e, desc_b, note)
            pu = convert_to_str(pu)
            pt = convert_to_str(pt)
            # inseriamo i valori nelle celle
            doc.insertTextIntoCell( table, "A{0}".format(i), tar)
            doc.insertTextIntoCell( table, "B{0}".format(i), text)
            doc.insertTextIntoCell( table, "C{0}".format(i), um)
            doc.insertTextIntoCell( table, "D{0}".format(i), quant)
            doc.insertTextIntoCell( table, "E{0}".format(i), "{0}".format(pu))
            doc.insertTextIntoCell( table, "F{0}".format(i), "{0}".format(pt))
            i+=1
        # inserire report o riepilogo strutturato
        if len(lista_rpt) > 0:
            # inserisce una riga bianca
            doc.writeLn()
            # creo una tabella per contenere gli articoli
            rows = len(lista_rpt)+1
            columns = 4
            table1 = doc.createTable(rows, columns)
            # eseguo il parsing degli elementi della lista e li dispongo nella tabella
            i=2
            for art in lista_rpt:
                tipo, cod, desc, pu, pt = art
                # inserimento di carattere speciale prima della stampa
                if tipo.startswith('$$$'): pass #FIXME testo in grassetto!!
                # preparazione del testo prima della stampa
                desc = desc.replace('\n', '')
                pu = convert_to_str(pu)
                pt = convert_to_str(pt)
                # inseriamo i valori nelle celle
                doc.insertTextIntoCell( table1, "A{0}".format(i), cod)
                doc.insertTextIntoCell( table1, "B{0}".format(i), desc)
                doc.insertTextIntoCell( table1, "C{0}".format(i), pu)
                doc.insertTextIntoCell( table1, "D{0}".format(i), pt)
                i+=1
                #doc.writeTabSeparated(cod, desc, pu, pt); doc.writeLn()
